package cz.cvut.fel.agents.pdv.evaluation.log;

import cz.cvut.fel.agents.pdv.dsand.DSProcess;
import cz.cvut.fel.agents.pdv.evaluation.Scenario;
import cz.cvut.fel.agents.pdv.evaluation.ScenarioBuilder;
import cz.cvut.fel.agents.pdv.raft.Client;
import cz.cvut.fel.agents.pdv.raft.RaftProcess;
import java.util.ArrayList;
import java.util.Comparator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.stream.Collectors;

/**
 * Scenar replikace logu.
 */
public class ScenarioForLogReplication extends Scenario<ClientCoordinatorForLogReplication> {

  /**
   * Vrati instanci stavitele pro nakonfigurovani a nasledne vytvoreni instance scenare
   */
  public static ScenarioBuilder<ScenarioForLogReplication> builder() {
    return new ScenarioBuilder<ScenarioForLogReplication>() {
      @Override
      public ScenarioForLogReplication build() {
        return new ScenarioForLogReplication(numberOfProcessesInCluster,
            reliabilityOfNetworkConnections, processToKill, restartProcess, reconnectProcess,
            expectedNumberOfProcessesInPartition, stepsOfSimulation, networkDelays, lengthOfFailure,
            spanBetweenFailures, countOfFailures, numberOfClients, requestPeriod,
            stopRequestBeforeEndOfSimulation, retryInterval, maxRequestsPerTick);
      }
    };
  }

  private ScenarioForLogReplication(int numberOfProcessesInCluster,
      double reliabilityOfNetworkConnections, int processToKill, boolean restartProcess,
      boolean reconnectProcess, int expectedNumberOfProcessesInPartition, long stepsOfSimulation,
      int networkDelays, int lengthOfFailure, int spanBetweenFailure, int countOfFailure,
      int numberOfClients, int requestPeriod, int stopRequestBeforeEndOfSimulation,
      int retryInterval, int maxRequestsPerTick) {
    super(() -> new ClientCoordinatorForLogReplication(retryInterval,
            stepsOfSimulation - stopRequestBeforeEndOfSimulation, requestPeriod,
            maxRequestsPerTick), numberOfProcessesInCluster, reliabilityOfNetworkConnections,
        processToKill, restartProcess, reconnectProcess, expectedNumberOfProcessesInPartition,
        stepsOfSimulation,
        networkDelays, lengthOfFailure, spanBetweenFailure, countOfFailure, numberOfClients);
  }

  @Override
  public boolean evaluateSimulation(List<Client<String>> clients,
      List<RaftProcess<Map<String, String>>> clusterProcesses,
      ClientCoordinatorForLogReplication clientCoordinator, List<Set<String>> partitions,
      Set<String> deadProcesses, boolean recoveredFromFailure) {

    // existuji nejake nevyrizene pozadavky? v tomto scenari pocitame s tim, ze vzdy zpracujete
    // vsechny pozadavky. V evaluaci na BRUTE, muze byt i pripad, kdy s timto pocitat nebudeme
    if (clientCoordinator.hasAnyUncommittedRequests()) {
      System.err.println("Nektere pozadavky nebyly zpracovany.");
      return false;
    }

    boolean correctResults = false;

    // system se zotavil ze selhani nebo k nemu ani nedoslo, data by mela byt konzistentni
    // v takovem pripade by vsechny logy mely byt shodne
    if (recoveredFromFailure) {

      // vybereme log nejakeho procesu
      String snapshotCheck = clusterProcesses.get(0).getLastSnapshotOfLog().get().get("A");

      // jsou vsechny snapshoty stejne
      if (clusterProcesses.stream()
          .map(RaftProcess::getLastSnapshotOfLog)
          .map(map -> map.get().get("A"))
          .allMatch(snapshotCheck::equals)) {
        correctResults = match(clientCoordinator, snapshotCheck);
      }
    } else {

      // nedoslo k zotaveni systemu. jeden spolecny lidr neni vubec zaruceny, proto
      // overime, ze nejvetsi skupina - podsit nebo zive procesy - ma stejny log

      //vsechny procesy, ktere nas zajimaji
      List<String> processes = !deadProcesses.isEmpty() ? clusterProcesses.stream()
          .map(DSProcess::getId)
          .filter(s -> !deadProcesses.contains(s))
          .collect(Collectors.toList()) : new ArrayList<>(partitions.stream()
          .max(Comparator.comparingInt(Set::size))
          .get());
      List<RaftProcess<Map<String, String>>> raftProcesses = clusterProcesses.stream()
          .filter(process -> processes.contains(process.getId()))
          .collect(Collectors.toList());

      // vybereme log nejakeho procesu
      String snapshotCheck = raftProcesses.get(0).getLastSnapshotOfLog().get().get("A");

      // jsou vsechny snapshoty stejne
      if (raftProcesses.stream()
          .map(RaftProcess::getLastSnapshotOfLog)
          .map(map -> map.get().get("A"))
          .allMatch(snapshotCheck::equals)) {
        correctResults = match(clientCoordinator, snapshotCheck);
      }
    }

    // vysledek evaluace
    if (correctResults) {
      System.out.println("Vsechny procesy maji stejny log.");
      return true;
    } else {
      System.err.println("Aspon dva propojene procesy maji ruzny log.");
      return false;
    }
  }

  /**
   * V metode zkontrolujeme, jestli bylo mozne vygenerovat snapshot, ktery obsahuji s procesy s
   * ohledem na zasialane pozadavky a poradi (kombinaci) jejich prijimani.
   * Je treba pocitat s ruznymi kombinacemi vysledku, nemame zarucene poradi
   * (ani pri vyzvedavani zprav z inboxu)
   */
  private boolean match(ClientCoordinatorForLogReplication clientCoordinator,
      String snapshotCheck) {
    int index = 0;
    for (int i = 0; i < clientCoordinator.getExpectedResults().size(); i++) {

      // znaky, ktere ocekavame v subsekvenci
      List<String> expectedSequence = new ArrayList<>(
          clientCoordinator.getExpectedResults().get(i));

      // cast podsekvence ke kontrole
      String sequence = snapshotCheck.substring(index, index + expectedSequence.size());

      // kontrola, zda jsou pritomny vsechny ocekavane znaky v subsekvenci
      for (String anExpectedSequence : expectedSequence) {
        int ind = sequence.indexOf(anExpectedSequence);
        if (ind >= 0) {
          StringBuilder sb = new StringBuilder(sequence);
          sb.deleteCharAt(ind);
          sequence = sb.toString();
        } else {
          return false;
        }
      }
      index = index + expectedSequence.size();
    }
    return true;
  }

}
